/*
 *	ldd.c
 *	list dynamic library dependencies
 *	for HP-UX 9.X
 *	see ld(1),a.out_800(4)
 */
static char *rcsid="$Id: ldd.c,v 1.4 1995/09/03 17:46:28 kabe Exp $";
static char *version="$VERSION: ldd-hp9-1.1 $";
#include <stdio.h>
#include <string.h>
#include <stdlib.h>	/*getenv*/
#include <sys/param.h>	/*MAXPATHLEN*/
#include <a.out.h>
#include <shl.h>

/* subspace name containig shared-library loading info*/
#define _SHLIB_INFO "$SHLIB_INFO$"

/* local prototypes */
void	ldd(char *filename);
void	dumpShlibInfo(FILE *a_out_file,long dl_header_loc);
char	*basename(char	*path);
char	*libBasename(char *basename);
size_t	fseekread(FILE *file,long int pos,void *ptr,size_t size,size_t nitems);
void	qerror(char *format,long arg1);
char	*Malloc(size_t size,size_t nitems);

int verbose = 0;
int showfilename = 1;

int main(argc,argv)
int	argc;
char	*argv[];
{
	int	c;
	extern int optind,optopt;
	char	*usage = "Usage: ldd [-v] filename ...";

	while((c=getopt(argc,argv,":v"))!=-1){
		switch(c){
		case 'v':	verbose=1; break;
		default:
			fprintf(stderr,"%s\n",usage);
			exit(1);
		}
	}
	if(optind==argc){ fprintf(stderr,"%s\n",usage); exit(1);}
	if(optind+1==argc)/*only one file*/showfilename=0;

	for( ; optind<argc; optind++){
		ldd(argv[optind]);
	}
	return 0;
}

void ldd(char *filename)
{
	FILE	*a_out_file;

	struct header header;	/* a.out header */
	static char	*space_strings = NULL;	/* strings table in the header */
	static SCNHDR *subspH = NULL;		/* subspace header list */
	int	i;
	char	*system;
	char	*type;

	if(NULL==(a_out_file = fopen(filename,"r"))){
		qerror("ldd: %s unreadable",(long)filename);
		return;
	}

	/**************** load a.out header ***/
	fread(&header,sizeof(struct header),1,a_out_file);

	switch(header.system_id){
	case HP9000S800_ID:
		system = " s800"; break;
	case _PA_RISC1_1_ID:
		system = " PA-RISC1.1"; break;
	case _PA_RISC1_2_ID:
		system = " PA-RISC1.2"; break;
	case _PA_RISC2_0_ID:
		system = " PA-RISC2.0"; break;
	default:
		goto not_exec;
	}
	switch(header.a_magic){
	case EXEC_MAGIC:
		type = " normal executable"; break;
	case SHARE_MAGIC:
		type = " shared executable"; break;
	case DEMAND_MAGIC:
		type = " demand-load executable"; break;
	case DL_MAGIC:
		type = " dynamic load library"; break;
	case SHL_MAGIC:
		type = " shared library"; break;
	default:
		goto not_exec;
	}
	/*printf("version_id = %d\n",header.version_id);*/

	(showfilename||verbose) && printf("%s: ",filename);
	(verbose) && printf("%s%s ",system,type);

	/********** read space_strings, the strings used in a.out header **/
	if(space_strings)	free(space_strings);
	space_strings = (char*)Malloc(header.space_strings_size,1);
	fseekread(a_out_file,header.space_strings_location,space_strings,1,header.space_strings_size);
	
	/********** read subspace headers **/
	if(subspH)	free(subspH);
	subspH = (SCNHDR*)Malloc(SCNHSZ,header.subspace_total);
	fseekread(a_out_file,header.subspace_location,subspH,SCNHSZ,header.subspace_total);

	/********** scan through subspace headers for $SHLIB_INFO$ **/
	for(i=0;i<header.subspace_total;i++){
		if(!strcmp(space_strings+subspH[i].STR_INDEX,_SHLIB_INFO)){
			/************* $SHLIB_INFO$ found **/
			/*printf("%s:\n",_SHLIB_INFO);*/
			if(subspH[i].initialization_length == 0L){
				/** initalize length of $SHLIB_INFO$ subspace was zero; thus no dynamic link information */
				printf("statically linked\n",filename);
			}else{
				(showfilename||verbose)&&printf("\n");
				dumpShlibInfo(a_out_file,subspH[i].file_loc_init_value);
			}
		}
	}
	fclose(a_out_file);
	return;

not_exec:
	fprintf(stderr,"ldd: %s: not an executable\n",filename);
	fclose(a_out_file);
	return;
}

void dumpShlibInfo(FILE *a_out_file,long dl_header_loc)
{
	struct dl_header dl_header;
	static struct shlib_list_entry *shlib_entries = NULL;
	static char	*dl_string_table = NULL;
	char	dlPath[2048];
	int	i;

	/************* load dl_header **/
	/** (this section seems to be
	 ** not documented in the man pages) **/
	fseekread(a_out_file,dl_header_loc+0L,&dl_header,sizeof(struct dl_header),1);
	if(dl_header.hdr_version != DL_HDR_VERSION_ID){
		printf("dl_header.hdr_version not DL_HDR_VERSION_ID\n");
		return;
	}
	/*printf("\t\thighest highwater mark = %d\n",dl_header.highwater_mark);*/

	/************* load shlib_list_entry-s **/
	if(shlib_entries)	free(shlib_entries);
	shlib_entries = (struct shlib_list_entry *)Malloc(sizeof(struct shlib_list_entry),dl_header.shlib_list_count);
	fseekread(a_out_file,dl_header_loc+dl_header.shlib_list_loc,shlib_entries,sizeof(struct shlib_list_entry),dl_header.shlib_list_count);
	/************* load string_table used in dl_header **/
	if(dl_string_table)	free(dl_string_table);
	dl_string_table = (char*)Malloc(dl_header.string_table_size,1);
	fseekread(a_out_file,dl_header_loc+dl_header.string_table_loc,dl_string_table,1,dl_header.string_table_size);

	/************* build library search path from ***/
	/*             embedded path and SHLIB_PATH to dlPath; see ld(1) **/
	dlPath[0]='\0';
	if(dl_header.flags&SHLIB_PATH_ENABLE && dl_header.flags&SHLIB_PATH_FIRST){
		strcat(dlPath,getenv("SHLIB_PATH"));
	}
	if(dl_header.flags&EMBED_PATH_ENABLE){
		strcat(dlPath,":");
		strcat(dlPath,dl_string_table+dl_header.embedded_path);
	}
	if(dl_header.flags&SHLIB_PATH_ENABLE && !(dl_header.flags&SHLIB_PATH_FIRST)){
		strcat(dlPath,":");
		strcat(dlPath,getenv("SHLIB_PATH"));
	}
	/* dlPath ~= ":/lib/pa1.1:/usr/lib/pa1.1:/lib:/usr/lib" */

	if(verbose){
		printf("\tembedded search path\t= \"%s\"\n",(dl_header.flags&EMBED_PATH_ENABLE)?dl_string_table+dl_header.embedded_path:"");
		printf("\tSHLIB_PATH order\t= %s\n",(dl_header.flags&SHLIB_PATH_ENABLE)?((dl_header.flags&SHLIB_PATH_FIRST)?"first":"last"):"disabled");
		printf("\tresulting search path\t= \"%s\"\n",dlPath);
	}

	/************* scan through shlib_entries **/
	/** usually first entry(i=0) is a.out itself, so start from i=1 */
	for(i= 1 ; i<dl_header.shlib_list_count; i++){
		char	*fullLibname = dl_string_table+shlib_entries[i].shlib_name;
		int	found = 0;

		if(shlib_entries[i].dash_l_reference){
	/** search list is only searched on "-lx" and "-l: x" linking; see ld(1) **/
			char 	*dpath;
			printf("\t-l%s ",libBasename(basename(fullLibname)));
			verbose && printf("(%s) ",fullLibname);
			printf("=> ");
			/** scan through dlPath **/
			for(dpath=strtok(dlPath,":"); dpath; dpath=strtok(NULL,":")){
				char	searchPath[MAXPATHLEN];
				/* null path "::" removed by strtok */
				sprintf(searchPath,"%s/%s",dpath,basename(fullLibname));
				if(0==access(searchPath,R_OK)){
					printf("%s\n",searchPath);
					found = 1;
				}
			}
			if(!found){
				if(0==access(fullLibname,R_OK)){
					printf("%s\n",fullLibname);
				}else{
					printf("(not found)\n");
				}
			}
		}else{/*full-path linked*/
	/** libraries linked with full-path are not searched **/
	/** only directly tested (see ld(1))**/
			printf("\t%s",fullLibname);
			if(0==access(fullLibname,R_OK)){
				;
			}else{
				printf(" not found\n");
			}
		}/*-l*/
	}

}

char	*basename(char	*path)
{
	char	*ret = strrchr(path,'/');
	if(ret==NULL)	return path;
	else	return ret+1;	/*skip '/'*/
}

char	*libBasename(char	*basename)
{
	static	char	libBase[MAXPATHLEN];
	char	*p;
	strcpy(libBase,basename+strlen("lib"));
	if(NULL!=(p=strchr(libBase,'.')))	*p='\0'; /* libxxx.sl */
	return libBase;
}

size_t fseekread(FILE *file,long int pos,void *ptr,size_t size,size_t nitems)
{
	if(fseek(file,pos,SEEK_SET)){
		qerror("ldd: fseek failed\n",NULL);
		exit(1);
	}
	return fread(ptr,size,nitems,file);
}

void qerror(char *format,long arg1)
{
	char	errbuf[1024];
	sprintf(errbuf,format,arg1);
	perror(errbuf);
}

char *Malloc(size_t size,size_t nitems)
{
	char	*ret = malloc(size*nitems);
	if(ret==NULL){
		fprintf(stderr,"ldd: malloc failed\n");
		exit(1);
	}
	return ret;
}
